#define HONEYDANBER /* means use ASCII pids in lock files */
#if defined(SHARE_DEBUG)
#define LOG_LOCKS
#endif
/*+-----------------------------------------------------------------------
	eculock.c -- lock file management
	wht@n4hgf.Mt-Park.GA.US

  Defined functions:
	check_utmp()
	create_lock_file(name)
	lock_tty()
	unlock_tty()

Lock files under SCO are supposed to use the direct line name
(lower-case last letter); we create only the lower-case case, but
check for both.  I have most definitely seen somebody create an
upper-case lock file (somebody named getty).  But that was on a middle-
aged XENIX revision.
------------------------------------------------------------------------*/
/*+:EDITS:*/
/*:08-25-1991-14:39-wht@n4hgf-SVR4 port thanks to aega84!lh */
/*:08-10-1991-17:39-wht@n4hgf-US_WEGOTIT handling */
/*:08-09-1991-11:07-wht@n4hgf-configurable lock directory */
/*:07-25-1991-12:56-wht@n4hgf-ECU release 3.10 */
/*:10-16-1990-20:43-wht@n4hgf-add SHARE_DEBUG */
/*:09-19-1990-19:36-wht@n4hgf-ecu_log_event now gets pid for log from caller */
/*:08-14-1990-20:40-wht@n4hgf-ecu3.00-flush old edit history */

#include "ecu.h"
#include "utmpstatus.h"

extern int errno;
extern char ungetty_ttyname[];
extern char lopen_err_str[];

/*+-------------------------------------------------------------------------
	check_utmp()
return 0 if line available, else LOPEN code
--------------------------------------------------------------------------*/
int
check_utmp()
{
register utstatus;
register status = 0;

	switch(utstatus = utmp_status(shm->Lline))
	{
		case US_DIALOUT:	/* enabled for login, currently dialout */
			status = LOPEN_DIALOUT_IN_USE;
			break;
		case US_LOGGEDIN:	/* enabled for login, in use */
			status = LOPEN_ENABLED_IN_USE;
			break;
		case US_NOTFOUND:	/* not in utmp, or getty dead */
			break;
		case US_WEGOTIT:	/* we own the line */
			status = LOPEN_WE_GOT_IT;	/* not really an error */
			break;
		case US_LOGIN:		/* enabled for login, idle */
			status = ungetty_get_line();
			break;
	}

#if defined(LOG_LOCKS)
{ char s64[64];
	sprintf(s64,"UTMPCHK %s st=%d ut=%d",shm->Lline,status,utstatus);
	ecu_log_event(getpid(),s64);
}
#endif

	return(status);

}	/* end of check_utmp */

/*+-----------------------------------------------------------------------
	void unlock_tty()
------------------------------------------------------------------------*/
void
unlock_tty()
{
	ungetty_return_line();
	if(LLCKname[0])
		unlink(LLCKname);
	LLCKname[0] = 0;

}	/* end of unlock_tty */

/*+-------------------------------------------------------------------------
	create_lock_file(name)
--------------------------------------------------------------------------*/
int
create_lock_file(name)
char *name;
{
register fd;
int pid = getpid();
char LTMP_fname[64];
#if defined(HONEYDANBER)
char pid10str[12];
#endif

	errno = 0;
#ifdef SVR4
	sprintf(LTMP_fname,"/var/spool/locks/LTMP.%d",pid);
#else
	strcpy(LTMP_fname,lock_dir_name);
	sprintf(LTMP_fname + strlen(LTMP_fname),"/LTMP.%05d",pid);
#endif
	if((fd = creat(LTMP_fname,0444)) < 0)
	{
		if(errno == EACCES)
		{
			strcpy(lopen_err_str,"lock error - try chmod 0777 ");
			strcat(lopen_err_str,lock_dir_name);
		}
		unlink(LTMP_fname);
		return(-1);
	}
#if defined(HONEYDANBER)
	sprintf(pid10str,"%10d\n",getpid());
	write(fd,pid10str,11);
#else
	write(fd,(char *)&pid,sizeof(int));
#endif

	chmod(LTMP_fname,0444);	/* some programs seem to think writable
							 * lock file is game for killing
							 */
	close(fd);

	fd = link(LTMP_fname,name);		/* use 'fd' for link return code */
	unlink(LTMP_fname);
	chmod(name,0444);

#if defined(LOG_LOCKS)
{ char s128[128];
  extern char *errno_text();
	sprintf(s128,"CRLOCK %s status=%d errno=%s",name,fd,errno_text(errno));
	ecu_log_event(getpid(),s128);
}
#endif

	return(fd);
}	/* end of create_lock_file */

/*+-------------------------------------------------------------------------
	lock_tty() - create lock files for tty name in 'shm->Lline'
--------------------------------------------------------------------------*/
lock_tty()
{
register itmp;

	if(itmp = make_lock_name(shm->Lline,LLCKname))
		return(itmp);

#ifndef SVR4
	if(itmp = check_utmp())
	{
		if(itmp == LOPEN_WE_GOT_IT)		/* if we already have the line ... */
			return(0);					/* ... no need to re-lock */
		if(itmp != LOPEN_ENABLED)		/* if other than enabled idle ... */
			return(itmp);				/* ... return the error */
	}
#endif

#if defined(GETTY_LOCKS_TTY)
	if(!ungetty_ttyname[0])	/* if getty did not lock line */
	{
#endif
		if(create_lock_file(LLCKname))
		{
			if(itmp = is_active_lock(LLCKname))
			{
				ungetty_return_line();
				errno = EACCES; /* for termecu() */
				return(itmp);
			}
			if(create_lock_file(LLCKname))
			{
				ungetty_return_line();
				errno = EACCES; /* for termecu() */
				return(LOPEN_LCKERR);
			}
		}
#if defined(GETTY_LOCKS_TTY)
	}
#endif

	return(0);
}	/* end of lock_tty */

/* end of eculock.c */
/* vi: set tabstop=4 shiftwidth=4: */
