/* #define NAP_DEBUG */
/*+-------------------------------------------------------------------------
	ecutty.c - local tty (console) functions
	wht@n4hgf.Mt-Park.GA.US

  Defined functions:
	_setcolor(clrs)
	B_to_timeout_msec(c_cflag,st_rdev)
	color_name_to_num(cname)
	get_initial_colors()
	get_ttymode()
	get_ttyname()
	ring_bell()
	setcolor(new_colors)
	setcolor_internal(ntokens,tokens)
	termio_to_kbd_chars()
	ttyflush(flush_type)
	ttygetc(xkey_ok)
	ttygets(str,maxsize,flags,delim)
	ttygets_esd(tesd,flags,append_flag)
	ttyinit(param)
	ttymode(arg)
	ttynormal_opost()

    In SCO versions, ECU keeps the the state of the normal and
    reverse video foreground and background colors in a 32-bit value:

     00000000001111111111222222222233
     01234567890123456789012345678901
     0000|--|0000|--|0000|--|0000|--|
          fg      bk      fg      bk
           reverse      normal

    The color values are per the SCO extended color definitons:

    black    0     gray         8
    blue     1     lt_blue      9
    green    2     lt_green    10
    cyan     3     lt_cyan     11
    red      4     lt_red      12
    magenta  5     lt_magenta  13
    brown    6     yellow      14
    white    7     hi_white    15

    With Lothar's ISC SVR4, the format is

     00000000001111111111222222222233
     01234567890123456789012345678901
     00000000000000000011????0100????
       0   0   0   0   3   f   4  o b
    where f is the foreground color
    and   b is the background color

    (I dont have one, so I'm guessing these colors are chosen from
    the ISO colors:

     BLACK       0
     RED         1
     GREEN       2
     YELLOW      3
     BLUE        4
     MAGENTA     5
     CYAN        6
     WHITE       7

--------------------------------------------------------------------------*/
/*+:EDITS:*/
/*:08-31-1991-13:29-wht@n4hgf2-look for colors in ECULIBDIR too */
/*:08-30-1991-04:12-wht@n4hgf2-restore colors wrong to do now if not SCO */
/*:08-30-1991-02:49-aega84!lh-use at_ansi.h/kd.h/CONS_GET under ISC SVR4 */
/*:08-25-1991-14:39-wht@n4hgf-SVR4 port thanks to aega84!lh */
/*:08-17-1991-18:29-wht@n4hgf-add kbdtest command */
/*:07-25-1991-12:57-wht@n4hgf-ECU release 3.10 */
/*:07-14-1991-18:18-wht@n4hgf-new ttygets functions */
/*:07-10-1991-16:19-wht@n4hgf-improve multi-char func key read timeout */
/*:03-20-1991-03:07-root@n4hgf-pts driver returns -1 on rdchk success! */
/*:03-19-1991-21:24-root@n4hgf-METRLINK_X11R4_PTS mods */
/*:01-29-1991-14:03-wht@n4hgf-more time for ESC vs fkey discrimination */
/*:01-29-1991-13:44-wht@n4hgf-load colors_normal w/ioctl GIO_ATTR if M_UNIX */
/*:12-01-1990-14:33-wht@n4hgf-more non-ansi - fkey mapping with nonansi.c */
/*:11-28-1990-15:56-wht@n4hgf-add non-ansi terminal support */
/*:08-14-1990-20:40-wht@n4hgf-ecu3.00-flush old edit history */

#include "ecu.h"
#include "esd.h"
#include "ecufkey.h"
#include "ecukey.h"
#include "ecuxkey.h"
#include "ecuerror.h"
#include "termecu.h"

/*
 * This is how we determine whether we are on a color terminal or not.
 * We handle SCO XENIX, UNIX , ISC 2.2 aqnd ISC SVR4.
 * If you have trouble with this add some kind of thing instead of MYSYS
 * and mail me the full details.  Search for MYSYS throughout here.
 */
#if defined(M_SYSV)
#   include <sys/machdep.h>
#else
#if defined(ISC) || defined(SVR4) /* && !defined(MYSYS) */
#  include <sys/at_ansi.h>
#  include <sys/kd.h>
#endif /* ISC */
#endif /* M_SYSV */

#define DEFINE_TTY_DATA
#include "ecutty.h"

/*
 * mapping table index to internal function code map
 *
 * the order of this table depends upon the pseudo-magic
 * KDEk_.. codes defined in ecufkey.h
 */
uchar KDEk_to_XF[] =
{
	XF1,		/* KDEk_F1 */
	XF2,		/* KDEk_F2 */
	XF3,		/* KDEk_F3 */
	XF4,		/* KDEk_F4 */
	XF5,		/* KDEk_F5 */
	XF6,		/* KDEk_F6 */
	XF7,		/* KDEk_F7 */
	XF8,		/* KDEk_F8 */
	XF9,		/* KDEk_F9 */
	XF10,		/* KDEk_F10 */
	XF11,		/* KDEk_F11 */
	XF12,		/* KDEk_F12 */
	XFcurup,	/* KDEk_CUU */
	XFcurdn,	/* KDEk_CUD */
	XFcurlf,	/* KDEk_CUL */
	XFcurrt,	/* KDEk_CUR */
	XFcur5,		/* KDEk_CU5 */
	XFpgup,		/* KDEk_PGUP */
	XFpgdn,		/* KDEk_PGDN */
	XFend,		/* KDEk_END */
	XFins,		/* KDEk_INS */
	XFbktab,	/* KDEk_BKTAB */
	XFhome		/* KDEk_HOME */
};

extern int hz;
extern uint tcap_LINES;
extern uint tcap_COLS;
extern int LINES;
extern int COLS;
extern char screen_dump_file_name[];

char *kde_text();

uint LINESxCOLS;
int current_ttymode = 0;
int ttymode_termecu_on_sigint = 0;
int tty_is_ansi;
int tty_is_pts;
int tty_is_pty;
int tty_is_multiscreen;
static int kbd_test_active = 0;

struct termio tty_termio_at_entry;
struct termio tty_termio_current;
struct stat tty_stat;
struct stat dn;
struct stat tty01;
struct stat ttyp0;
#if defined(METRLINK_X11R4_PTS)
struct stat pts000;
#endif /* METRLINK_X11R4_PTS */
struct stat console;

uchar kbdeof;			/* current input EOF */
uchar kbdeol2;			/* current secondary input EOL */
uchar kbdeol;			/* current input EOL */
uchar kbderase;			/* current input ERASE */
uchar kbdintr;			/* current input INTR */
uchar kbdkill;			/* current input KILL */
uchar kbdquit;			/* current input QUIT */
int echo_erase_char;	/* save users ECHOE bit */
int echo_kill_char;		/* save users ECHOK bit */
char kbd_is_7bit;		/* keyboard has parity */
long tty_escape_timeout = 40L;		/* timeout on waiting for char after ESC */

uchar *dole_out_tgc_accum = (uchar *)0;
int dole_out_tgc_accum_count = 0;

/*
 * color definitions per format described at top of source
 * we handle SCO XENIX and UNIX and ISC SVR4 but not "MYSYS"
 * (search for MYSYS near top of source for clue what MYSYS means)
 *
 * As of this writing, I don't know about colors on ISC 2.2.
 * but they might be the same as for SVR4.
 *
 */
#if defined(SVR4)
ulong colors_current = 0x00003740;
ulong colors_normal = 0x00003740;	/* */
ulong colors_initial = 0x00003740;	/* */
ulong colors_success = 0x00003740;	/* */
ulong colors_alert = 0x00003740L;	/* */
ulong colors_error = 0x00003140L;	/* */
ulong colors_notify = 0x00003340L;	/* */
#else	/* SCO XENIX and UNIX ---- other? */
ulong colors_current = 0x04070A00L;
ulong colors_normal = 0x04070A00L;	/* default lt_green/black red/white */
ulong colors_initial = 0x04070A00L;	/* default initial colors */
ulong colors_success = 0x07000A00L;	/* lt_green/black red/white */
ulong colors_alert = 0x0E000E00L;	/* yellow */
ulong colors_error = 0x04000400L;	/* red */
ulong colors_notify = 0x08000800L;	/* gray */
#endif

int use_colors = 0;		/* set by ttyinit, but default no */
int tty_not_char_special;

/*+-------------------------------------------------------------------------
	B_to_timeout_msec(c_cflag,st_rdev) - CBAUD code to ESC timeout msec
--------------------------------------------------------------------------*/
/*ARGSUSED*/
ulong
B_to_timeout_msec(c_cflag,st_rdev)
ushort c_cflag;
ushort st_rdev;
{
long ms = 300L;

	/* make pts sweat, but don't make as many mistakes */
#if defined(METRLINK_X11R4_PTS)
	if(tty_is_pts)
		return(ms);
#endif /* METRLINK_X11R4_PTS */

	/* make network/xterm/pty sweat, but don't make as many mistakes */
	if(tty_is_pty)
		return(ms);

	/* if multiscreen, 3 ticks is pu-lenty */
	if(tty_is_multiscreen)
		return((long)(1000/hz * 3));

	/* baud rate fiddling */
	switch(c_cflag & CBAUD)
	{
		/*       char times * time/char */
		case B110:	ms = 10 * 100;
		case B300:	ms = 10 * 33;
		case B600:	ms = 10 * 16;
		case B1200:	ms = 10 * 8;
		case B2400:	ms = 10 * 4;
		default:		/* many character times for packetized ... */
			ms = 400L;	/* ... modems used for console */
	}
	return(ms);

}	/* end of B_to_timeout_msec */

/*+-------------------------------------------------------------------------
	color_name_to_num(cname)
--------------------------------------------------------------------------*/
int
color_name_to_num(cname)
char *cname;
{
register COLOR *color = colors;
register itmp;

	while(color->name)
	{
		if((itmp = strcmp(color->name,cname)) > 0)
			return(-1);
		if(!itmp)
			return(color->num);
		color++;
	}
	return(-1);

}	/* end of color_name_to_num */

/*+-------------------------------------------------------------------------
	_setcolor(clrs)
--------------------------------------------------------------------------*/
void
_setcolor(clrs)
ulong clrs;
{
#ifdef SVR4
	char fg[4], bg[4];
#endif

	if(!use_colors || tty_not_char_special)
		return;

#ifdef SVR4
	/* normal */
	sprintf(fg,"%s", (clrs >> 8) & 0xFF);
	sprintf(bg,"%s", clrs & 0xFF);
	ff(se,"\033[%sm", fg);
	ff(se,"\033[%sm", bg);
#else /* SCO */
	/* normal */
	ff(se,"\033[=%ldF\033[=%ldG",(clrs >> 8) & 0xFF,clrs & 0xFF);

	/* reverse */
	ff(se,"\033[=%ldH\033[=%ldI",(clrs >> 24) & 0xFF,(clrs >> 16) & 0xFF);
#endif
	colors_current = clrs;
}	/* end of _setcolor */

/*+-------------------------------------------------------------------------
	setcolor(new_colors)

requires termcap init to have been done
--------------------------------------------------------------------------*/
void
setcolor(new_colors)
ulong new_colors;
{
	if(tty_not_char_special)
		return;

	if(!use_colors)
	{
		if((new_colors == colors_notify) || (new_colors == colors_alert) ||
		   (new_colors == colors_error))
		{
			tcap_stand_out();
		}
		else
			tcap_stand_end();
		return;
	}
	_setcolor(new_colors);
	tcap_stand_end();
}	/* end of setcolor */

/*+-------------------------------------------------------------------------
	setcolor_internal(ntokens,tokens)

returns 0 on success, else token number in error + 1
--------------------------------------------------------------------------*/
int
setcolor_internal(ntokens,tokens)
int ntokens;
char **tokens;
{
ulong fgnd;
ulong bgnd;

	if(tty_not_char_special || !use_colors)
		return(0);

	if(ntokens == 2)
		tokens[2] = "black";

	if((fgnd = (ulong)color_name_to_num(tokens[1])) > 15)
		return(2);
	if((bgnd = (ulong)color_name_to_num(tokens[2])) > 15) 
		return(3);

	if(!strcmp(tokens[0],"normal"))
	{
		colors_normal &= 0xFFFF0000L;
		colors_normal |= (fgnd << 8) | bgnd;
		setcolor(colors_normal);
	}
	else if(!strcmp(tokens[0],"reverse"))
	{
		colors_normal &= 0x0000FFFFL;
		colors_normal |= (fgnd << 24) | (bgnd << 16);
		setcolor(colors_normal);
	}
	else if(!strcmp(tokens[0],"notify"))
		colors_notify = (fgnd << 24) | (bgnd << 16) | (fgnd << 8) | bgnd;
	else if(!strcmp(tokens[0],"success"))
		colors_success = (fgnd << 24) | (bgnd << 16) | (fgnd << 8) | bgnd;
	else if(!strcmp(tokens[0],"alert"))
		colors_alert = (fgnd << 24) | (bgnd << 16) | (fgnd << 8) | bgnd;
	else if(!strcmp(tokens[0],"error"))
		colors_error = (fgnd << 24) | (bgnd << 16) | (fgnd << 8) | bgnd;
	else
		return(1);

	return(0);

}	/* end of setcolor_internal */

/*+-------------------------------------------------------------------------
	restore_initial_colors() - under UNIX, restore color scheme at execution

Under non-SCO-S5R3, don't have initial color, so use colors_normal
--------------------------------------------------------------------------*/
void
restore_initial_colors()
{
#if defined(M_UNIX) || defined(SVR4)
	if(use_colors)
		setcolor(colors_initial);
	else
#endif
	{
		tcap_blink_off();
		tcap_bold_off();
		tcap_underscore_off();
	}
}	/* end of restore_initial_colors */

/*+-------------------------------------------------------------------------
	get_initial_colors() - read colors at time of execution from driver

     00000000001111111111222222222233
     01234567890123456789012345678901
     0000|--|0000|--|0000|--|0000|--|
          fg      bk      fg      bk
           reverse      normal

  Thanks for the G2 on this to staceyc@sco.COM (Stacey Campbell)
--------------------------------------------------------------------------*/
#if defined(M_UNIX)
void
get_initial_colors()
{
#if !defined(GIO_ATTR)
#define GIO_ATTR  ('a' << 8) | 0	/* Ioctl call for current attribute */
#endif
	uint cur_attr;
	ulong fgnd;
	ulong bgnd;

	colors_normal = 0L;

/*
 * first, reverse, so we can end up with normal colors selected
 */
	write(1,"\033[7m",4);		/* select reverse */
	cur_attr = (uint)ioctl(0, GIO_ATTR, 0);
	fgnd = (ulong)cur_attr & 0x0F;
	bgnd = (ulong) (cur_attr >> 4) & 0x0F;
	colors_normal |= (fgnd << 24) | (bgnd << 16);

/*
 * now, normal
 */
	write(1,"\033[m",3);		/* select normal */
	cur_attr = (uint)ioctl(0, GIO_ATTR, 0);
	fgnd = (ulong)cur_attr & 0x0F;
	bgnd = (ulong) (cur_attr >> 4) & 0x0F;
	colors_normal |= (fgnd << 8) | bgnd;
	colors_initial = colors_normal;		/* save for restore_initial_colors */

}	/* end of get_initial_colors */
#endif

/*+-------------------------------------------------------------------------
	read_colors_file()
--------------------------------------------------------------------------*/
void
read_colors_file()
{
FILE *fp;
char s128[128];
#define MAX_COLOR_TOKENS 6
char *tokens[MAX_COLOR_TOKENS];
int ntokens;
char *cptr;
int itmp;

	if(tty_not_char_special)
		return;

#if defined(M_UNIX)
	get_initial_colors();
#endif

	get_home_dir(s128);
	strcat(s128,"/.ecu/colors");
	if(access(s128,4))
	{
		get_home_dir(eculibdir);
		strcat(s128,"/colors");
		if(access(s128,4))
			return;
	}

	fp = fopen(s128,"r");

	while(fgets(s128,sizeof(s128),fp))
	{
		if(s128[0] == '#')			/* comment? */
			continue;
		if(itmp = strlen(s128))		/* itmp = len; if > 0 ... */
		{
			itmp--;
			s128[itmp] = 0;			/* ... strip trailing NL */
		}
		cptr = s128;				/* first call to str_token, -> buff */
		while((*cptr == 0x20) || (*cptr == TAB))
			cptr++;				/* strip leading spaces */
		if(*cptr == 0)				/* if line all blank, skip it */
			continue;

		build_str_array(s128,tokens,MAX_COLOR_TOKENS,&ntokens);
		if(ntokens < 2)
			continue;

		setcolor_internal(ntokens,tokens);

	}			/* while records left to ready */
	fclose(fp);
}	/* end of read_colors_file */

/*+-------------------------------------------------------------------------
	ring_bell()
--------------------------------------------------------------------------*/
void
ring_bell()
{
	if(tty_not_char_special)
		return;

	fputc(7,se);
}	/* end of ring_bell */

/*+-------------------------------------------------------------------------
	termio_to_kbd_chars()
--------------------------------------------------------------------------*/
void
termio_to_kbd_chars()
{
	kbdintr =  (tty_termio_at_entry.c_cc[VINTR])
		? (tty_termio_at_entry.c_cc[VINTR]  & 0x7F) : '\377';
	kbdquit =  (tty_termio_at_entry.c_cc[VQUIT])
		? (tty_termio_at_entry.c_cc[VQUIT]  & 0x7F) : '\377';
	kbderase = (tty_termio_at_entry.c_cc[VERASE])
		? (tty_termio_at_entry.c_cc[VERASE] & 0x7F) : '\377';
	kbdkill =  (tty_termio_at_entry.c_cc[VKILL])
		? (tty_termio_at_entry.c_cc[VKILL]  & 0x7F) : '\377';
	kbdeof =   (tty_termio_at_entry.c_cc[VEOF])
		? (tty_termio_at_entry.c_cc[VEOF]   & 0x7F) : '\04';
	kbdeol2 =  (tty_termio_at_entry.c_cc[VEOL])
		? (tty_termio_at_entry.c_cc[VEOL]   & 0x7F) : '\377';
	kbdeol =   (tty_termio_at_entry.c_iflag & ICRNL)
		? '\r' : '\n';

	kbd_is_7bit = ((tty_termio_at_entry.c_cflag & PARENB) != 0);
	echo_erase_char = tty_termio_at_entry.c_lflag & ECHOE;
	echo_kill_char = tty_termio_at_entry.c_lflag & ECHOK;

}	/* end of termio_to_kbd_chars */

/*+-------------------------------------------------------------------------
	ttyinit(param)
--------------------------------------------------------------------------*/
void
ttyinit(param)
uchar param;
{
int itmp;
#if defined(M_SYSV) || defined(ISC) || defined(SVR4) /* && !defined(MYSYS) */
int monitor_type;
#endif
char *cptr;
int fddevtty = open("/dev/tty",O_RDONLY,0);

	/*
	 * get control tty control chars in case stdin not tty
	 */
	if(fddevtty >= 0)
	{
		ioctl(fddevtty,TCGETA,(char *)&tty_termio_at_entry);
		close(fddevtty);
		termio_to_kbd_chars();
	}

	sigint = 0;			/* see xmtr signal handlers */

	memset((char *)&tty_stat,0xFF,sizeof(struct stat));
	memset((char *)&ttyp0,0xFF,sizeof(struct stat));
	memset((char *)&console,0xFF,sizeof(struct stat));
	stat("/dev/console",&console);
	stat("/dev/null",&dn);
	stat("/dev/tty01",&tty01);
	stat("/dev/ttyp0",&ttyp0);

/*
 * if stdin not open or is /dev/null or is non-character-device
 */

	itmp = fstat(TTYIN,&tty_stat);
	if(itmp || ((tty_stat.st_mode & S_IFMT) != S_IFCHR) ||
		((dn.st_ino == tty_stat.st_ino) && (dn.st_rdev == tty_stat.st_rdev)))
	{
		tcap_LINES = LINES = 25;	/* fake necessary termcap/curses vars */
		tcap_COLS = COLS = 80;
		LINESxCOLS = tcap_LINES * tcap_COLS;
		shm->scr_lines = tcap_LINES;
		shm->scr_cols = tcap_COLS;
		shm->scr_size = LINESxCOLS;

		tty_not_char_special = 1;
		tty_is_ansi = (param == TTYINIT_FORCE_ANSI);
		tty_is_multiscreen = 0;
		return;
	}

/*
 * if pty
 */
	if((tty_stat.st_rdev & 0xFF00) == (ttyp0.st_rdev & 0xFF00))
		tty_is_pty = 1;

/*
 * if pts
 */

#if defined(METRLINK_X11R4_PTS)
	memset((char *)&pts000,0xFF,sizeof(struct stat));
	stat("/dev/pts000",&pts000);
	if((tty_stat.st_rdev & 0xFF00) == (pts000.st_rdev & 0xFF00))
		tty_is_pts = 1;
#endif /* METRLINK_X11R4_PTS */

	/*
	 * use color if we are on a multiscreen and video supports it
	 * also, remember whether or not we are on a multiscreen
	 * (I ain't proud of this beyond being a valiant attempt)
	 */
	use_colors = 0;
#if defined(M_SYSV) || defined(ISC) || defined(SVR4) /* && !defined(MYSYS) */
	if( ((itmp = ioctl(TTYIN,CONS_GET,&monitor_type)) >= 0) &&
		(use_colors = (monitor_type != MONO)))
	{
		read_colors_file();
		setcolor(colors_normal);
	}
	tty_is_multiscreen = !(itmp < 0);
#endif

	/*
	 * save initial tty state
	 */
	ioctl(TTYIN,TCGETA,(char *)&tty_termio_at_entry);
	tty_escape_timeout =
		B_to_timeout_msec(tty_termio_at_entry.c_cflag,tty_stat.st_rdev);

	termio_to_kbd_chars();

	tty_termio_current = tty_termio_at_entry;
	current_ttymode = 0;

	get_home_dir(screen_dump_file_name);
	strcat(screen_dump_file_name,"/.ecu/screen.dump");

#if defined(M_SYSV)
	cptr = (char *)0;
	if(param)
		tty_is_ansi = (param == TTYINIT_FORCE_ANSI);
	else
		tty_is_ansi = ((cptr = getenv("TERM")) && (ulindex(cptr,"ansi") != -1));
#else
	tty_is_ansi = 0;	/* if not SCO, always use ~/.ecu/nonansikeys */
	cptr = getenv("TERM");
#endif

	if(!tty_is_ansi && cptr)
		nonansi_key_read(cptr);

/* initialize termcap */
	tcap_init();			/* read termcap strings */

/* yetch - magic number gretching for lines and columns */
	if((tcap_LINES < 16) || (tcap_LINES > 43))
	{
		ff(se,"screen height must be >= 16 and <= 43 lines. (found %dx%d)\r\n",
			tcap_COLS,tcap_LINES);
		termecu(TERMECU_USAGE);
	}
	if(tcap_COLS != 80)
	{
		ff(se,"terminal width must be 80 columns.\r\n");
		termecu(TERMECU_USAGE);
	}
	if(!tcap_LINES || !tcap_COLS)
	{
		tcap_LINES = 25;
		tcap_COLS = 80;
	}
	if(tcap_LINES > 43)
		tcap_LINES = 43;
	if(tcap_COLS > 80)
		tcap_COLS = 80;
	LINESxCOLS = tcap_LINES * tcap_COLS;
	shm->scr_lines = tcap_LINES;
	shm->scr_cols = tcap_COLS;
	shm->scr_size = LINESxCOLS;

}	/* end of ttyinit */

/*+-----------------------------------------------------------------------
	ttymode(arg) -- control user console (kbd/screen)

  Where arg ==
	0 restore attributes saved at start of execution
	1 raw mode (send xon/xoff, but do not respond to it, no ISIG/SIGINT)
	2 raw mode (same as 1 but allow keyboard interrupts)
	3 same as 2 but terminate program on SIGINT

------------------------------------------------------------------------*/
void
ttymode(arg)
int arg;
{

	if(tty_not_char_special)
		return;

	ttymode_termecu_on_sigint = (arg == 4);

	switch(arg)
	{
	case 0:
		ioctl(TTYIN,TCSETAW,(char *)&tty_termio_at_entry);
		tty_termio_current = tty_termio_at_entry;
		current_ttymode = 0;
		break;

	case 1:
	case 2:
	case 3:
		tty_termio_current = tty_termio_at_entry;

		tty_termio_current.c_cflag &= ~(PARENB | PARODD);
		tty_termio_current.c_cflag |= CS8;

		/* don't want to honor tty xon/xoff, but pass to other end */
		tty_termio_current.c_iflag &=
			~(INLCR | ICRNL | IGNCR | IXON | IUCLC | ISTRIP);
		tty_termio_current.c_iflag |= IXOFF;	/* this end will xon/xoff */

		tty_termio_current.c_oflag |= OPOST;
		tty_termio_current.c_oflag &= ~(OLCUC | ONLCR | OCRNL | ONOCR | ONLRET);

		tty_termio_current.c_lflag &= ~(ICANON | ISIG | ECHO);
		if(arg > 1)
			tty_termio_current.c_lflag |= ISIG;

		tty_termio_current.c_cc[VMIN] = 1;
		tty_termio_current.c_cc[VTIME] = 0;

		ioctl(TTYIN,TCSETAW,(char *)&tty_termio_current);
		current_ttymode = arg;
		break;

	default:
		ff(se,"\r\nttymode: invalid argument %d\r\n",arg);
		break;
	}
}	/* end of ttymode */

/*+-------------------------------------------------------------------------
	ttynormal_opost()
--------------------------------------------------------------------------*/
void
ttynormal_opost()
{
	tty_termio_current.c_oflag = tty_termio_at_entry.c_oflag;
	ioctl(TTYIN,TCSETAW,(char *)&tty_termio_current);

}	/* end of ttynormal_opost */

/*+-------------------------------------------------------------------------
	int	get_ttymode()
--------------------------------------------------------------------------*/
int
get_ttymode()
{
	return(current_ttymode);
}	/* end of get_ttymode */

/*+-----------------------------------------------------------------------
	ttyflush(flush_type) -- flush tty driver input &/or output buffers

0 == input buffer
1 == output buffer
2 == both buffers
------------------------------------------------------------------------*/
void
ttyflush(flush_type)
int flush_type;
{
	if(tty_not_char_special)
		return;

	ioctl(TTYIN,TCXONC,(char *)0); /* stop tty output */

#if !defined(M_I286)
	ioctl(TTYIN,TCFLSH,(char *)flush_type);
#else
	/* avoid 286 compiler warning of cast int to far ptr */
	switch(flush_type)
	{
		case 0:
			ioctl(TTYIN,TCFLSH,(char *)0); break;
		case 1:
			ioctl(TTYIN,TCFLSH,(char *)1); break;
		case 2:
			ioctl(TTYIN,TCFLSH,(char *)2); break;
	}
#endif

	ioctl(TTYIN,TCXONC,(char *)1);	/* restart tty output */

	if(flush_type != 1)
	{
		dole_out_tgc_accum = (uchar *)0;
		dole_out_tgc_accum_count = 0;
	}

}	/* end of ttyflush */

/*+-------------------------------------------------------------------------
	ttyrdchk() - see if key pressed and not read
--------------------------------------------------------------------------*/
int
ttyrdchk()
{
	return(rdchk(TTYIN) || dole_out_tgc_accum_count);
}	/* end of ttyrdchk */

/*+-------------------------------------------------------------------------
	ttygetc(xkey_ok) -- get a key from the keyboard
if UNIX or XENIX, map extended keys to sign-bit-set special value
if xkey_ok is 0, disallow extended keys
--------------------------------------------------------------------------*/
uint
ttygetc(xkey_ok)
int xkey_ok;
{
uchar ctmp;
extern int errno;
register uint itmp = 0;
long timeout_remaining;
static uchar tgc_accum[16];
uchar map_nonansi_key();

	if(tty_not_char_special)	/* this really is unexplored territory */
	{
		ctmp = 255;
		read(0,(char *)&ctmp,1);
		return((uint)ctmp);
	}

	if(dole_out_tgc_accum_count)
	{
		ctmp = *dole_out_tgc_accum++;
		dole_out_tgc_accum_count--;
		if(kbd_is_7bit)
			ctmp &= 0x7F;
		return((uint)ctmp);
	}

GET_KEY:
	errno = 0;
	if(read(TTYIN,(char *)&ctmp,1) < 0)
	{
		if(errno == EINTR)
			goto GET_KEY;
		perror_errmsg("keyboard");
		termecu(TERMECU_TTYIN_READ_ERROR);
	}

	/* if escape from SCO/"ansi" terminal */
	if(tty_is_ansi && ((ctmp & 0x7F) == ESC))
	{
		itmp = 0;
		timeout_remaining = tty_escape_timeout;
#if defined(NAP_DEBUG)
		ff(se,"timeout_remaining = %ld hzmsec=%ld\r\n",
			timeout_remaining,hzmsec);
#endif
		while((!isalpha(ctmp)) && (itmp < sizeof(tgc_accum) - 1) &&
			(timeout_remaining > 0))
		{
			timeout_remaining -= Nap(hzmsec);
#if defined(NAP_DEBUG)
			ff(se,"timeout_remaining = %ld\r\n",timeout_remaining);
#endif
			if(!rdchk(TTYIN))
				continue;
			read(TTYIN,(char *)&ctmp,1);
			tgc_accum[itmp++] = ctmp;
			if(itmp == 2)
				break;
		}
		tgc_accum[itmp] = 0;
		if(kbd_test_active)
		{
			char title[32];
			sprintf(title,"--> func key '%s' (%d key codes received)",
				kde_text(ctmp),itmp);
			hex_dump(tgc_accum,-itmp,title,1);
		}
		if(!itmp)				/* no subsequent chars, so ... */
			return(ESC);		/* return the escape */
		else if((itmp == 2) && (tgc_accum[0] == '['))
		{
			switch(tgc_accum[1] | 0x80)
			{
				case XFcur5:
					screen_dump(screen_dump_file_name);
					goto GET_KEY;
				case XFcurup: case XFcurdn: case XFcurrt: case XFcurlf:
				case XFend: case XFpgdn: case XFhome: case XFpgup: case XFins:
				case XF1: case XF2: case XF3: case XF4: case XF5: case XF6:
				case XF7: case XF8: case XF9: case XF10: case XF11: case XF12:
				case XFbktab:
					if(xkey_ok)
						return((uint)tgc_accum[1] | 0x80);
					/* fall thru -- xkey not allowed */
				default:
					ring_bell();
					goto GET_KEY;
			}
			/*NOTREACHED*/
		}
		/* not func key -- must be typamatic control key */
		if(kbd_test_active)
		{
			char title[32];
			if(itmp > 1)
			{
				sprintf(title,
					"--> no func key recognized (%d key codes received)",itmp);
				hex_dump(tgc_accum,-itmp,title,1);
			}
		}
		dole_out_tgc_accum_count = itmp - 1;
		dole_out_tgc_accum = tgc_accum + 1;
		if(kbd_is_7bit)
			tgc_accum[0] &= 0x7F;
		return((uint)tgc_accum[0]);
	}
	else if(!tty_is_ansi &&			/* this is getting crazy */
			((ctmp >= 0x01) && (ctmp <= 0x1F) ||
			 (ctmp >= 0x81) && (ctmp <= 0x9F)) &&
			(ctmp != kbderase) && (ctmp != kbdkill) &&
			(ctmp != kbdeol) && (ctmp != kbdeol2) &&
			(ctmp != kbdintr) && (ctmp != kbdeof) )
	{
		tgc_accum[0] = ctmp;
		tgc_accum[itmp = 1] = 0;
		timeout_remaining = tty_escape_timeout;
#if defined(NAP_DEBUG)
		ff(se,"timeout_remaining = %ld hzmsec=%ld\r\n",
			timeout_remaining,hzmsec);
#endif
		while(((ctmp = map_nonansi_key(tgc_accum,itmp)) >= XF_no_way) &&
			(timeout_remaining > 0))
		{
			timeout_remaining -= Nap(hzmsec);
#if defined(NAP_DEBUG)
			ff(se,"timeout_remaining = %ld\r\n",timeout_remaining);
#endif
			if(!rdchk(TTYIN))
				continue;
			read(TTYIN,(char *)&ctmp,1);
			if(itmp == (sizeof(tgc_accum) - 1))	/* do not allow overflow */
			{
				ctmp = XF_no_way;
				break;
			}
			timeout_remaining = tty_escape_timeout;
			tgc_accum[itmp++] = ctmp;
		}
		tgc_accum[itmp] = 0;
		if((ctmp == XF_not_yet) && (itmp == 1))
		{
			if(kbd_is_7bit)
				tgc_accum[0] &= 0x7F;
			return((uint)tgc_accum[0]);
		}
		else if(ctmp < XF_no_way)	/* if we got a map */
		{
			if(kbd_test_active)
			{
				char title[32];
				sprintf(title,"--> func key '%s' (%d key codes received)",
					kde_text(ctmp),itmp);
				hex_dump(tgc_accum,-itmp,title,1);
			}
			if(!xkey_ok)
			{
				ring_bell();
				goto GET_KEY;
			}
			switch(ctmp)
			{
				case KDEk_CU5:
					screen_dump(screen_dump_file_name);
					goto GET_KEY;
				default:
					return((uint)KDEk_to_XF[ctmp]);
			}
			/*NOTREACHED*/
		}
		/* not func key -- must be typamatic control key */
		if(kbd_test_active)
		{
			char title[32];
			if(itmp > 1)
			{
				sprintf(title,
					"--> no func key recognized (%d key codes received)",itmp);
				hex_dump(tgc_accum,-itmp,title,1);
			}
		}
		dole_out_tgc_accum_count = itmp - 1;
		dole_out_tgc_accum = tgc_accum + 1;
		if(kbd_is_7bit)
			tgc_accum[0] &= 0x7F;
		return((uint)tgc_accum[0]);
	}

	/*
	 * simple key, not special
	 */
	if(kbd_is_7bit)
		ctmp &= 0x7F;
	return((uint)ctmp);
}	/* end if ttygetc */

/*+-----------------------------------------------------------------------
	ttygets(str,maxsize,flags,delim,pstrpos)

flags & TG_CRLF   - echo cr/lf terminator
flags & TG_XDELIM - extended delimiter set
                    (Home, End, PgUp, PgDn, CurUp, CurDn)
flags & TG_EDIT   - redisplay/edit current string
flags & TG_IPOS   - if edit, use initial string pos
------------------------------------------------------------------------*/
void
ttygets(str,maxsize,flags,delim,pstrpos)
register char *str;
int maxsize;
int flags;
uchar *delim;
int *pstrpos;
{
register inch;
register strcount = 0;
register strpos = 0;
int insert_mode = 0;

	--maxsize;		/* decrement for safety */

	if(flags & TG_EDIT)
	{
		strpos = strcount = strlen(str);
		fputs(str,se);
		if(pstrpos && (*pstrpos > 0) && (*pstrpos <= strcount))
			strpos = *pstrpos;
		tcap_curleft(strcount - strpos);
	}

	while(1)
	{
		inch = ttygetc(1);
		*delim = (uchar)inch;	/* last char will always be the delimiter */
		if((inch == kbdintr) || (inch == ESC))
		{
			tcap_curright(strcount - strpos);
			while(strcount)
			{
				fputc(BS,se);
				fputc(SPACE,se);
				fputc(BS,se);
				strcount--;
			}
			str[strcount] = 0;
			*delim = ESC;
			goto RETURN;
		}
		else if(inch == kbdkill)
		{
			tcap_curright(strcount - strpos);
			while(strcount)
			{
				fputc(BS,se);
				fputc(SPACE,se);
				fputc(BS,se);
				strcount--;
			}
			strpos = 0;
			*str = 0;
			continue;
		}
		else if(inch == kbderase)
		{
			if(strcount)
			{
				if(strcount == strpos)
				{
					fputc(BS,se);
					fputc(SPACE,se);
					fputc(BS,se);
					strcount--,strpos--;
				}
				else
				{
					if(!strpos)
						continue;
					mem_cpy(str + strpos - 1,str + strpos,strcount - strpos);
					fputc(BS,se);
					str[--strcount] = 0;
					strpos--;
					fputs(str + strpos,se);
					fputc(' ',se);
					tcap_curleft(strcount - strpos + 1);
				}
			}
			str[strcount] = 0;
			continue;
		}
		else if(inch == XFins)
		{
			insert_mode = !insert_mode;
			continue;
		}
		else if(inch == XFcurlf)
		{
			if(strpos)
			{
				strpos--;
				tcap_curleft(1);
			}
			continue;
		}
		else if(inch == XFcurrt)
		{
			if(strpos < strcount)
			{
				strpos++;
				tcap_curright(1);
			}
			continue;
		}

		if(flags & TG_XDELIM)	/* extended delimiter */
		{
			switch(inch)
			{
				case XFhome:
				case XFend:
				case XFpgup:
				case XFpgdn:
				case XFcurup:
				case XFcurdn:
#ifdef notdef
					tcap_curright(strcount - strpos);
					while(strcount)
					{
						fputc(BS,se);
						fputc(SPACE,se);
						fputc(BS,se);
						strcount--;
					}
#endif
					str[strcount] = 0;
					goto RETURN;
			}
		}

		switch(inch)
		{
			case CRET:
				*delim = NL;
			case NL:
				str[strcount] = 0;
				tcap_curright(strcount - strpos);
				if((flags & TG_CRLF))
					ff(se,"\r\n");
				goto RETURN;

			case CTL_L:
			case CTL_R:
				tcap_curright(strcount - strpos);
				ff(se,"%s (insert mode %s)\r\n",make_char_graphic(inch,0),
					(insert_mode) ? "ON" : "OFF");
				tcap_eeol();
				fputs(str,se);
				tcap_curleft(strcount - strpos);
				break;

			default:
				if((inch < SPACE) || (inch >= 0x7F))
				{
					ring_bell();
					break;
				}
				if(strpos == strcount)
				{
					if(strcount == maxsize)
					{
						ring_bell();
						continue;
					}
					str[strcount++] = inch;
					strpos++;
					fputc(inch,se);
				}
				else
				{
					if(insert_mode)
					{
						if(strcount == maxsize)
						{
							ring_bell();
							continue;
						}
						mem_cpy(str+strpos+1,str+strpos,strcount-strpos);
						str[strpos] = inch;
						strcount++;
						str[strcount] = 0;
						fputs(str + strpos++,se);
						tcap_curleft(strcount - strpos);
					}
					else
					{
						str[strpos++] = inch;
						fputc(inch,se);
					}
				}
				str[strcount] = 0;
				continue;
		}
	}

RETURN:
	if(pstrpos)
		*pstrpos = strpos;

}	/* end of ttygets() */

#ifdef notdef
void
ttygets(str,maxsize,flags,delim,pstrpos)
register char *str;
int maxsize;
int flags;
uchar *delim;
int *pstrpos;
{
ff(se,"\r\n>>>ttygets str='%s' strpos=%d flag=%d\r\n",str,*pstrpos,flags);
	_ttygets(str,maxsize,flags,delim,pstrpos);
ff(se,"\r\n>>>ttygets buf='%s' strpos=%d delim=%02x\r\n",str,*pstrpos,*delim);
}
#endif

/*+-------------------------------------------------------------------------
	ttygets_esd(tesd,flags,append_flag)
--------------------------------------------------------------------------*/
ttygets_esd(tesd,flags,append_flag)
ESD *tesd;
int flags;
int append_flag;
{
char *pb = tesd->pb;
int maxcb = tesd->maxcb;
uchar delim;

	if(append_flag)
	{
		pb += tesd->cb;
		maxcb -= tesd->cb;
	}
	else
	{
		pb = tesd->pb;
		maxcb = tesd->maxcb;
		tesd->cb = 0;
	}

	ttygets(pb,maxcb,flags,&delim,(int *)0);

	if(delim == ESC)
	{
		if(!append_flag)
			esdzero(tesd);
		return(eProcAttn_ESCAPE);
	}

	tesd->cb = strlen(tesd->pb);
	plogs(pb);
	if(flags & 1)
		plogc(NL);
	return(0);

}	/* end of ttygets_esd */

/*+-------------------------------------------------------------------------
	kbd_test() - test keyboard handler
--------------------------------------------------------------------------*/
void
kbd_test()
{
uint ctmp = 0;

	pputs("Press keys to test (ESCape to exit)\n");
	kbd_test_active = 1;
	while(ctmp != ESC)
	{
		ctmp = ttygetc(1);
		
		if((ctmp < 0x80) && dole_out_tgc_accum_count)
		{
			pprintf("    got %d key sequence %s ",
				dole_out_tgc_accum_count,hex_to_ascii_name(ctmp));
			while(dole_out_tgc_accum_count)
			{
				pprintf("%s ",hex_to_ascii_name(*dole_out_tgc_accum++));
				dole_out_tgc_accum_count--;
			}
			pputs("\n");
			ctmp = 0;
			continue;
		}
		pputs("    got ");
		if(ctmp >= 0x80)
			pprintf("fkey '%s'\n",xf_text(ctmp));
		else
			pprintf("key  '%s'\n",hex_to_ascii_name(ctmp));
	}
	kbd_test_active = 0;
	pputs("keyboard test complete\n\n");
	
}	/* end of kbd_test */

/*+-------------------------------------------------------------------------
	char *get_ttyname() - return pointer to static string

This routine is largely a crock and is likely to explode at any rev or twist
--------------------------------------------------------------------------*/
char *
get_ttyname()
{
#ifndef OLD_WAY
char *ttyname();
	return(ttyname(TTYIN));
#else
static char ttname[64];
register unsigned int rdev;
register char *cptr;

	if(tty_not_char_special)
		return("stdin");
	else if(!tty_is_multiscreen)
		return("non-multiscreen");

	rdev = (unsigned)tty_stat.st_rdev;
	if(rdev == 0x0301)
		strcpy(ttname,"/dev/console");
#if defined(M_UNIX)
	else if(rdev == 0x0000)
		strcpy(ttname,"/dev/syscon");
#endif
	else
	{
		strcpy(ttname,"/dev/tty");
		cptr = ttname + 8;

		if(rdev < 0x000C)
		{
			*cptr++ = '0' + ((rdev + 1) / 10);
			*cptr++ = '0' + ((rdev + 1) % 10);
		}
		else if(!(rdev & ~0x58F))
		{
			*cptr++ = (rdev & 0x0008) ? '2' : '1';
			*cptr++ = ((rdev & 0x0080) ? 'A' : 'a') + (rdev & 0x0007);
		}
		else
		{
			*cptr++ = '?';
			*cptr++ = '?';
		}
		*cptr = 0;
	}

	return(ttname);
#endif
}	/* end of get_ttyname */

/* end of ecutty.c */
/* vi: set tabstop=4 shiftwidth=4: */
