/* CHK=0x2F4F */
char *revision = "1.02";
/*+-----------------------------------------------------------------------
	kbdtest3.c -- test keyboard values
	wht@n4hgf.Mt-Park.GA.US

  See ecu manual section titled Function Key Recognition

  Defined functions:
	dump_putc(ch)
	dump_puts(str)
	hex_dump16(int16)
	hex_dump32(int32)
	hex_dump4(int4)
	hex_dump8(int8)
	hex_dump_fp(fp,str,len,title,terse_flag)
	k3ttymode(arg)
	main(argc,argv)
	read_kbd_string(buf,max)
	termio_parity_text(cflag)
	tputstrs(strs)
	write_nonansi_desc(fp,buf,buflen,name)
	xtoasc(ch,incl_3char)

------------------------------------------------------------------------*/
/*+:EDITS:*/
/*:08-28-1991-14:07-wht@n4hgf2-SVR4 cleanup by aega84!lh */
/*:08-17-1991-13:58-root@n4hgf-make kbd entirely raw */
/*:08-06-1991-13:12-wht@n4hgf-add parity reporting */
/*:08-06-1991-13:12-wht@n4hgf-some terminals reinvent parity bit's use */
/*:07-25-1991-12:58-wht@n4hgf-ECU release 3.10 */
/*:04-29-1991-18:24-wht@n4hgf-let us see what keyboards say */

#include <stdio.h>
#include <signal.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/errno.h>
#include <sys/stat.h>
#include <string.h>
#include <termio.h>

char *ctime();
char *getenv();

#define VMIN_VALUE  32
#define VTIME_VALUE 2

#define ff	fprintf
#define se	stderr

#define TTYIN   0				/* mnemonic */
#define TTYOUT  1				/* mnemonic */
#define TTYERR  2				/* mnemonic */

struct termio lv;		/* attributes for the line to remote */
struct termio tv0;		/* for saving, changing TTY atributes */
struct termio tv;		/* for saving, changing TTY atributes */

char *strs_intro[] = {
	"\n\n\n\n\n\n\n\n\n\n",
	"Let's learn your keyboard.  I'll be making a file named kbdtest3.out\n",
	"For each of the key names shown below, please press the key and wait.\n",
	"\n",
	"If you do not have a key on your keyboard, make an alternate choice\n",
	"or press the space bar if nothing seems reasonable.\n",
	"\n",
	"If you press a key but don't see further activity after a second or two\n",
	"press the slash '/' key unless you can choose a reasonable alternate.\n",
	"\n",
	"Do not use the same key for more than one function.\n",
	"\n",
	(char *)0
};

char *strs_thanks[] = {
	"\n",
	"Thank you.  If you wish to mail me the contents of kbdtest3.out,\n",
	"please include  a detailed description of the system and software\n",
	"(i.e., \"Metrolink xterm keyboard on SCO 3.2r2\")\n",
	"I also need the curses terminal type (TERM=\?\?\?) entry for this keyboard\n",
	"(I WANT you to mail me results for non-SCO/non-ISC-console keyboards.)\n",
	"If you had to hack this program, mail it in its entirety as well.\n",
	"\n",
	"My address: wht@n4hgf.Mt-Park.GA.US or emory!n4hgf!wht\n",
	(char *)0
};

struct keyname {
	char *ecuname;
	char *peoplename;
};

struct keyname need_names[] = {
	{ "BkTab",	"Back Tab (Shift Tab)" },
	{ "CU5",	"Unshifted Keypad 5" },
	{ "F1",		"F1" },
	{ "F2",		"F2" },
	{ "F3",		"F3" },
	{ "F4",		"F4" },
	{ "F5",		"F5" },
	{ "F6",		"F6" },
	{ "F7",		"F7" },
	{ "F8",		"F8" },
	{ "F9",		"F9" },
	{ "F10",	"F10" },
	{ "F11",	"F11" },
	{ "F12",	"F12" },
	{ "Ins",	"Ins" },
	{ "Home",	"Home" },
	{ "End",	"End" },
	{ "PgUp",	"PgUp" },
	{ "PgDn",	"PgDn" },
	{ "CUU",	"Cursor Up" },
	{ "CUD",	"Cursor Down" },
	{ "CUL",	"Cursor Left" },
	{ "CUR",	"Cursor Right" },
	{(char *)0,(char *)0}
};

char *parity_text = "<undetermined>";
static FILE *dumpfp;

/*+-------------------------------------------------------------------------
	dump_putc(ch)
--------------------------------------------------------------------------*/
void
dump_putc(ch)
char ch;
{
	fputc(ch,dumpfp);
}	/* end of dump_putc */


/*+-------------------------------------------------------------------------
	dump_puts(str)
--------------------------------------------------------------------------*/
void
dump_puts(str)
char *str;
{
	fputs(str,dumpfp);
}	/* end of dump_puts */

/*+-----------------------------------------------------------------------
	hex_dump#... subservient routines
------------------------------------------------------------------------*/
void hex_dump4(int4)
unsigned char int4;
{
	int4 &= 15;
	dump_putc((int4 >= 10) ? (int4 + 'A' - 10) : (int4 + '0'));
}

void hex_dump8(int8)
unsigned char int8;
{
	hex_dump4(int8 >> 4);
	hex_dump4(int8);
}

void hex_dump16(int16)
unsigned short int16;
{
	hex_dump8(int16 >> 8);
	hex_dump8(int16);
}

void hex_dump32(int32)
unsigned long int32;
{
	hex_dump16(int32 >> 16);
	hex_dump16(int32);
}

/*+-----------------------------------------------------------------
	hex_dump_fp(fp,str,len,title,terse_flag)

  if 'title' not NULL, title is printed... 'terse_flag'
  controls whether or not the title is "conspicuous" with
  hyphens before and after it making title line >70 chars long
------------------------------------------------------------------*/
void
hex_dump_fp(fp,str,len,title,terse_flag)
FILE *fp;
char *str;
int len;
char *title;
int terse_flag;
{
int istr;
register ipos = 0;
register itmp;

	dumpfp = fp;

	if(title && (istr = strlen(title)))
	{
		if(!terse_flag)
		{
			ipos = (73 - istr) / 2;
			itmp = ipos;
			while(itmp--)
				dump_putc('-');
			dump_putc(' ');
			if(istr & 1)
				ipos--;
		}
		dump_puts(title);
		if(!terse_flag)
		{
			dump_putc(' ');
			while(ipos--)
				dump_putc('-');
		}
		if(dumpfp == se)
			dump_puts("\r\n");
		else
			dump_puts("\n");

	}

	istr = 0;
	while(istr < len)
	{
		hex_dump16(istr);
		dump_putc(' ');
		for(itmp = 0; itmp < 16; ++itmp)
		{
			ipos = istr + itmp;
			if(ipos >= len)
			{
				if(!terse_flag)
					dump_puts("   ");
				continue;
			}
			dump_putc(' ');
			hex_dump8(str[ipos]);
		}
		dump_puts(" | ");
		for(itmp = 0; itmp < 16; ++itmp)
		{
			ipos = istr + itmp;
			if( (ipos) >= len)
			{
				if(!terse_flag)
					dump_putc(' ');
			}
			else
			{
				dump_putc((str[ipos] >= ' ' && str[ipos] < 0x7f)
		 			? str[ipos] : '.' );
			}
		}
		if(dumpfp == se)
			dump_puts(" |\r\n");
		else
			dump_puts(" |\n");
		istr += 16;
	}   /* end of while(istr < len) */

}	/* end of hex_dump_fp */

/*+-----------------------------------------------------------------------
	xtoasc(character) - Make all chars "printable"

  returns pointer to a static string containing printable version
  of a character.  If control char, printed as "^A", etc.
  if incl_3char set true, then space + ASCII assignment (e.g. "NUL") is
  appended to the string for non-printable graphics
------------------------------------------------------------------------*/
char *
xtoasc(ch,incl_3char)
register unsigned char ch;
int incl_3char;
{
static char gg[8];
char *ascii_ctlstr =
"nulsohstxetxeotenqackbelbs ht nl vt ff cr so si dledc1dc2dc3dc4naksynetbcanem subescfs gs rs us sp ";

	if(ch == 0x7F)
		strcpy(gg,"del");
	else if(ch == 0x9b)
		strcpy(gg,"csi");
	else if(ch > 0x7F)
		sprintf(gg,"0x%02x",(unsigned char)ch);
	else if(ch > 0x20)
	{
		gg[0] = ch;
		gg[1] = 0;
	}
	else
	{
		strncpy(gg,ascii_ctlstr + (ch * 3),3);
		gg[3] = 0;
	}
	return(gg);
}	/* end of xtoasc */

/*+-------------------------------------------------------------------------
	write_nonansi_desc(fp,buf,buflen,name)
--------------------------------------------------------------------------*/
void
write_nonansi_desc(fp,buf,buflen,name)
FILE *fp;
unsigned char *buf;
int buflen;
char *name;
{
char s256[256];

	sprintf(s256,"    %s:%s:",name,name);
	while(strlen(s256) < (unsigned)20)
		strcat(s256," ");

	while(buflen--)
	{
		strcat(s256,xtoasc(*buf++));
		if(buflen)
			strcat(s256," ");
	}
	strcat(s256,"\n");

	fputs(s256,fp);

}	/* end of write_nonansi_desc */

/*+-------------------------------------------------------------------------
	tputstrs(strs)
--------------------------------------------------------------------------*/
void
tputstrs(strs)
char **strs;
{
	while(*strs)
		fputs(*strs++,stdout);
}	/* end of tputstrs */

/*+-----------------------------------------------------------------------
	k3ttymode(arg) -- control user console (kbd/screen)

  Where arg ==
	0 restore attributes saved at start of execution
	1 raw mode 

------------------------------------------------------------------------*/
void k3ttymode(arg)
int arg;
{
	if(arg)
	{
		(void)ioctl(TTYIN,TCGETA,&tv);
		tv.c_cflag &= ~(CS8 | PARENB | PARODD);
		tv.c_cflag |= CS8;
		tv.c_iflag &= ~(INLCR | ICRNL | IGNCR | IXOFF | IUCLC | ISTRIP);
		tv.c_lflag &= ~(ICANON | ISIG | ECHO);
		tv.c_cc[VEOF] = '\01';
		tv.c_cc[VEOL] = '\0';
		tv.c_cc[VMIN] = VMIN_VALUE;
		tv.c_cc[VTIME] = VTIME_VALUE;
		(void)ioctl(TTYIN,TCSETAW,&tv);
	}
	else
		(void)ioctl(TTYIN,TCSETAW,&tv0);
}

/*+-------------------------------------------------------------------------
	read_kbd_string(buf,max)
--------------------------------------------------------------------------*/
int
read_kbd_string(buf,max)
unsigned char *buf;
int max;
{
int count = read(TTYIN,buf,max);

#ifdef notdef
int itmp;
	for(itmp = 0; itmp < count; itmp++)
		buf[itmp] &= 0x7F;
#endif
	return(count);

}	/* end of read_kbd_string */

/*+-------------------------------------------------------------------------
	termio_parity_text(cflag)
--------------------------------------------------------------------------*/
char *
termio_parity_text(cflag)
unsigned short cflag;
{
	 return((cflag & PARENB) ? ((cflag & PARODD) ? "odd" : "even") : "none");
}	/* end of termio_parity_text */

/*+-----------------------------------------------------------------------
	main()
------------------------------------------------------------------------*/
main(argc,argv)
int argc;
char **argv;
{
int itmp;
int count;
struct keyname *need = need_names;
unsigned char instr[VMIN_VALUE + 1];
char s128[128];
FILE *fpout;
long now;
int errflg = 0;
int hex_out = 0;
char *outfile = "kbdtest3.out";
/* extern char *optarg; */
extern int optind;

	setbuf(stdout,NULL);
	setbuf(stderr,NULL);

	while((itmp = getopt(argc,argv,"hx")) != -1)
	{
		switch(itmp)
		{
			case 'h':
			case 'x':
				hex_out = 1;
				break;
			case '?':
				errflg++;
		}
	}

	if(optind == (argc - 1))
		outfile = argv[optind++];

	if(errflg || (optind != argc))
	{
		(void)fprintf(stderr,"usage: %s [-hx] [outfile]\n");
		exit(1);
	}

	tputstrs(strs_intro);
	if(!(fpout = fopen(outfile,"a")))
	{
		perror(outfile);
		exit(1);
	}

	ioctl(TTYIN,TCGETA,&tv0);		/* get original status */
	parity_text = termio_parity_text(tv0.c_cflag);

	if(!hex_out)
	{
	char *ttype;
		if(!(ttype = getenv("TERM")))
			ttype = "??";
		time(&now);
		fprintf(fpout,"# nonansikey for '%s' under ",ttype);
#if defined(M_SYSV)
		fputs("SCO\n",fpout);
#else
#if defined(ISC)
		fputs("ISC\n",fpout);
#else
#if defined(sun)
		fputs("SunOS\n",fpout);
#else
		fputs("??? OS\n",fpout);
#endif /* sun */
#endif /* ISC */
#endif /* M_SYSV */

		fprintf(fpout,"# built by kbdtest3 %s %s",revision,ctime(&now));
		fprintf(fpout,"# keyboard parity required = %s\n",parity_text);
#if defined(sun)
#ifdef notdef /* system() uses sockets $#@$^!&%# */
		fprintf(fpout,"# stty everything at kbdtest3 execution time:\n");
		fclose(fpout);
		strcpy(s128,"/usr/bin/stty everything | ");
		strcat(s128,"/usr/bin/sed -e 's/^/# /' >> ");
		strcat(s128,outfile);
#endif
#else
		fprintf(fpout,"# stty -a at kbdtest3 execution time:\n");
		fclose(fpout);
		strcpy(s128,"/bin/stty -a | /bin/sed -e 's/^/# /' >> ");
		strcat(s128,outfile);
#endif
		system(s128);
		if(!(fpout = fopen(outfile,"a")))
		{
			perror(outfile);
			exit(1);
		}
		fprintf(fpout,"%s\n",ttype);
	}

	printf("Your keyboard driver parity is set to %s\n",parity_text);
	printf("press ^D (0x04) to terminate program early\n\n");
	k3ttymode(1);

	while(need->ecuname)
	{
		printf("%-20.20s: ",need->peoplename);
		count = read_kbd_string(instr,VMIN_VALUE);
		if(!count)
		{
			printf("whoops ..... zero length read\n");
			break;
		}
		if(!count)
		{
			perror("keyboard");
			break;
		}

		for(itmp = 0; itmp < count; itmp++)
			printf("%02x ",instr[itmp]);
		printf("\n");

		if(hex_out)
			hex_dump_fp(fpout,instr,count,need->peoplename,0);
		else
		{
			if((count == 1) && (instr[0] == '/'))
			{
				fprintf(fpout,"# %s: dead key and no reasonable alternate\n",
					need->ecuname);
			}
			else if((count == 1) && (instr[0] == ' '))
			{
				fprintf(fpout,"# %s: no key and no reasonable alternate\n",
					need->ecuname);
			}
			else
			{
				write_nonansi_desc(fpout,(unsigned char *)instr,count,
					need->ecuname);
			}
		}

		if((count == 1) && ((instr[0] & 0x7F) == 4))
		{
			fputs("#Something is wrong with above entry.\n",fpout);
			break;
		}

		need++;
	}

	fputs("\n",fpout);
	fclose(fpout);
	k3ttymode(0);
	if(!need->ecuname)
	{
		printf(
		"\nRemember to set keyboard parity to \"%s\" when using this entry.\n",
		parity_text);
		tputstrs(strs_thanks);
	}
	else
	{
		printf("Something went wrong. Please try again.\n");
	}
	exit(0);

}	/* end of main */
/* vi: set tabstop=4 shiftwidth=4: */
