#define NONANSI_DEBUG
/*+-----------------------------------------------------------------
	nonansikey.c - keyboard function key -> ECU internal
	wht@n4hgf.Mt-Park.GA.US

  Defined functions:
	define_nonansi_key(bufptr)
	map_nonansi_key(buf,buflen)
	nonansi_key_init()
	nonansi_key_read(name)

------------------------------------------------------------------*/
/*+:EDITS:*/
/*:08-28-1991-14:51-wht@n4hgf2-look correctly for nonansikeys in ECULIBDIR */
/*:08-26-1991-05:45-wht@n4hgf2-# got included in key def */
/*:08-06-1991-13:19-wht@n4hgf-allow any code as first in key sequence */
/*:08-03-1991-14:44-wht@n4hgf-look for nonansikeys in ECULIBDIR too */
/*:07-25-1991-12:58-wht@n4hgf-ECU release 3.10 */
/*:03-20-1991-03:06-root@n4hgf-no Metro Link problems here */
/*:03-20-1991-01:04-root@n4hgf-diagnose Metro Link xterm differences */
/*:01-10-1991-23:15-wht@n4hgf-string overflow rptd by spooley@compulink.co.uk */
/*:12-01-1990-12:51-wht@n4hgf-creation, borrowing from and using ecufkey.c */

#include "ecu.h"
#include "ecukey.h"
#include "ecufkey.h"
#include "ecuxkey.h"
#include "ecufork.h"

char *keyset_idstr();

extern int tty_not_char_special;

KDE nonansi_key_table[KDE_COUNT];
char nonansi_key_name[32] = "";
#if defined(NONANSI_DEBUG)
static FILE *nadbg = (FILE *)0;
#endif

/*+-------------------------------------------------------------------------
	nonansi_key_init()
--------------------------------------------------------------------------*/
void
nonansi_key_init()
{
register itmp;
register KDE *tkde;
char *make_char_graphic();

	for(itmp = 0; itmp < KDE_COUNT; itmp++)
	{
		tkde = &nonansi_key_table[itmp];
		tkde->logical[0] = 0;
		tkde->count = 0;
		tkde->KDEt = (uchar)itmp;
	}

	nonansi_key_name[0] = 0;

}	/* end of nonansi_key_init */

/*+-------------------------------------------------------------------------
	define_nonansi_key(bufptr)

  returns 0 good keydef
         -1 if syntax error
--------------------------------------------------------------------------*/
int
define_nonansi_key(bufptr)
register char *bufptr;
{
register itmp;
register token_number = 0;
KDE *tkde = (KDE *)0;
int KDEt = 0;
int octothorpe = 0;
char *token_separator = ":";
char *token;
char *arg_token();
char *str_token();
char *skip_ld_break();

	while(!octothorpe &&
	   (token = (token_number < 2) ? str_token(bufptr,":")
	                               : arg_token(bufptr," \t")))
	{
		bufptr = (char *)0;	/* further calls to arg_token need NULL */
		token = skip_ld_break(token);
		strip_trail_break(token);

		switch(token_number)
		{
			case 0:		/* first field is key identifier */
				if((KDEt = keyset_idnum(token)) < 0)
				{
					ff(se,"  %s is not a legal key identifier\r\n",token);
					return(-1);
				}
				tkde = &nonansi_key_table[KDEt];
				tkde->logical[0] = 0;
				tkde->count = 0;
				break;

			case 1:		/* second field is logical key name */
				if(*token == '#')
					goto MISSING_LABEL;
				strncpy(tkde->logical,token,sizeof(tkde->logical));
				tkde->logical[sizeof(tkde->logical) - 1] = 0;
				token_separator = " \t"; /* space is tok sep now */
				break;

			case 2:		/* third field is first token of sequence */
				if(*token == '#')
					goto MISSING_SEQUENCE;

			default:	/* third and subsequent to define key */
				if(*token == '#')
				{
					octothorpe = 1;
					break;
				}
				if(tkde->count == sizeof(tkde->str))
				{
					ff(se,"  %s: output count too long",
						keyset_idstr(KDEt));
					return(-1);
				}
				if((itmp = ascii_to_hex(token)) < 0)
				{
					ff(se,"  %s: '%s' invalid\r\n",
						keyset_idstr(KDEt),token);
					return(-1);
				}
				tkde->str[tkde->count] = itmp;
				tkde->count++;
				break;
		}	/* end of switch(token_number) */

		if(octothorpe)
			break;

		token_number++;

	}		/* end while not end of record */

	switch(token_number)
	{
		case 0:
			ff(se,"define nonansi key logic error\r\n");
			termecu(TERMECU_XMTR_FATAL_ERROR);
			break;
		case 1:
MISSING_LABEL:
			ff(se,"%s: missing key label\r\n",keyset_idstr(KDEt));
			break;
		case 2:
MISSING_SEQUENCE:
			ff(se,"%s: missing char sequence\r\n",keyset_idstr(KDEt));
			break;
		default:
			/* special init string entry */
			if(KDEt == KDEk_InitStr)
				fwrite(tkde->str,1,tkde->count,se);
#if defined(NONANSI_DEBUG)
			else
				hex_dump_fp(nadbg,tkde->str,tkde->count,keyset_idstr(KDEt),0);
#endif
			return(0);
	}

	return(-1);		/* error */

}	/* end of define_nonansi_key */

/*+-------------------------------------------------------------------------
	nonansi_key_read(name) - read key-sequence-to-fkey map from nonansikeys
--------------------------------------------------------------------------*/
void
nonansi_key_read(name)
char *name;
{
register itmp;
char buf[128];
FILE *fp_keys;
int errstat = 0;
static char ecukeys_name[128];
#if defined(NONANSI_DEBUG)
	if(!nadbg)
		nadbg = fopen("/tmp/nadbg.log","w");
#endif

	nonansi_key_init();	/* clear any previous key defns */

	if(!ecukeys_name[0])
	{
		get_home_dir(ecukeys_name);
		strcat(ecukeys_name,"/.ecu/nonansikeys");
	}

	if(!(fp_keys = fopen(ecukeys_name,"r")))
	{
		strcpy(ecukeys_name,eculibdir);
		strcat(ecukeys_name,"/nonansikeys");
		if(!(fp_keys = fopen(ecukeys_name,"r")))
		{
			ff(se,"'nonansikeys' not in ~/.ecu or %s; unable to proceed\r\n",
				eculibdir);
			termecu(TERMECU_CONFIG_ERROR);
		}
	}

/* find nonansi_key name */
	errstat = 1;
	while((itmp = kde_fgets(buf,sizeof(buf),fp_keys)) != KDEt_EOF)
	{
		if((itmp == KDEt_COMMENT) || (itmp == KDEt_ENTRY))
			continue;
		if(!strcmp(buf,name))
		{
			errstat = 0;		/* indicate success */
			break;
		}
	}
	if(errstat)
	{
		ff(se,"terminal type '%s'\r\nnot found in %s; unable to proceed\r\n",
			name,ecukeys_name);
		termecu(TERMECU_CONFIG_ERROR);
	}

/* read past any other nonansi_key names matching this set */
	errstat = 1;
	while((itmp = kde_fgets(buf,sizeof(buf),fp_keys)) != KDEt_EOF)
	{
		if(itmp == KDEt_ENTRY)
		{
			errstat = 0;		/* indicate success */
			break;
		}
	}
	if(errstat)
	{
		ff(se,"terminal type '%s' has null entry in %s; unable to proceed\r\n",
			name,ecukeys_name);
		termecu(TERMECU_CONFIG_ERROR);
	}

/* we found the definition ... process it */
	errstat = 0;
	itmp = KDEt_ENTRY;
	do {
		if(itmp == KDEt_NAME)
			break;
		else if(itmp == KDEt_ENTRY)
		{
			if(define_nonansi_key(buf))
				errstat = 1;
		}
	} while((itmp = kde_fgets(buf,sizeof(buf),fp_keys)) != KDEt_EOF);

/* close up shop */
	strncpy(nonansi_key_name,name,sizeof(nonansi_key_name));
	nonansi_key_name[sizeof(nonansi_key_name) - 1] = 0;
	fclose(fp_keys);

	if(!nonansi_key_table[KDEk_HOME].count)
	{
		ff(se,"You MUST have a 'Home' key defined\r\n");
		errstat = 2;
	}
	if(!nonansi_key_table[KDEk_END].count)
	{
		ff(se,"You MUST have a 'End' key defined\r\n");
		errstat = 2;
	}
	if((errstat == 2) || (errstat && tty_not_char_special))
		termecu(TERMECU_CONFIG_ERROR);

	if(errstat)
	{
		ff(se,
			"Warning: key definition syntax errors\r\nContinue anyway (y,n)? ");
		if((itmp = ttygetc(0)) == 'Y' || (itmp == 'y'))
		{
			ff(se,"YES\r\n");
			return;
		}
		ff(se,"NO\r\n");
		termecu(TERMECU_CONFIG_ERROR);
	}

	/*NOTREACHED*/

}	/* end of nonansi_key_read */

/*+-------------------------------------------------------------------------
	map_nonansi_key(buf,buflen) - map char sequence to KDEt code

return KDEt code or XF_not_yet if no match yet, XF_no_way if no match possible
--------------------------------------------------------------------------*/
uchar
map_nonansi_key(buf,buflen)
uchar *buf;
int buflen;
{
register ikde;
register KDE *tkde;
int err_rtn = XF_no_way;

	if(!buflen)
		return(XF_not_yet);

#if defined(NONANSI_DEBUG)
	if(nadbg)
		hex_dump_fp(nadbg,buf,-buflen,"mapna",1);
#endif

	for(ikde = 0,tkde = nonansi_key_table; ikde < KDE_COUNT;
		ikde++,tkde++)
	{
		if((tkde->count == buflen) && !memcmp(tkde->str,buf,buflen))
		{
#if defined(NONANSI_DEBUG)
			if(nadbg)
				ff(nadbg,"--> %s\n",tkde->logical);
#endif
			return(tkde->KDEt);
		}
		if((tkde->count > buflen) &&
			(*(tkde->str + buflen) == *(buf + buflen)))
		{
			err_rtn = XF_not_yet;
		}
	}
	return(err_rtn);
}	/* end of map_nonansi_key */

/* end of nonansikey.c */
/* vi: set tabstop=4 shiftwidth=4: */
